#include <propeller.h>

#include "Library.h"

#define I2C_SCL         28
#define I2C_SDA         29

#define I2C_ACK         0
#define I2C_NAK         1
#define I2C_WRITE       0
#define I2C_READ        1

#define EEPROM_ADDR     0xA0

#define HIGH_EEPROM_OFFSET(a)  ((uint32_t)(a) - 0xc0000000 + 0x8000)

extern uint32_t _ovly_addr;
extern uint32_t _ovly_table[][2];

void pinMode(int pin, int mode)
{
    if (mode == INPUT) {
        DIRA &= ~(1 << pin);
    }
    else {
        DIRA |= 1 << pin;
    }
}

void digitalWrite(int pin, int value)
{
    if (value == LOW) {
        OUTA &= ~(1 << pin);
    }
    else {
        OUTA |= 1 << pin;
    }
}

int digitalRead(int pin)
{
    return (INA & (1 << pin)) >> pin;
}

void delay(unsigned int ms)
{
    waitcnt(CNT + ms * (_CLKFREQ / 1000));
}

void delayMicroseconds(unsigned int us)
{
    waitcnt(CNT + us * (_CLKFREQ / 1000000));
}

void shiftOut(int32_t dataPin, int32_t clockPin, int32_t bitOrder, uint32_t val)
{
    dataPin = (1 << dataPin);
    clockPin = (1 << clockPin);

    if (bitOrder == MSBFIRST) {
        for (int i = 0x80; i != 0; i >>= 1) {
            if ((val & i) == 0) {
                OUTA &= ~dataPin;
            }
            else {
                OUTA |= dataPin;
            }
            OUTA |= clockPin;
            OUTA &= ~clockPin;
        }
    }
    else {
        for (int i = 0x01; i != 0x100; i <<= 1) {
            if ((val & i) == 0) {
                OUTA &= ~dataPin;
            }
            else {
                OUTA |= dataPin;
            }
            OUTA |= clockPin;
            OUTA &= ~clockPin;
        }
    }
}

uint32_t shiftIn(int32_t dataPin, int32_t clockPin, int32_t bitOrder)
{
    int i;
    uint32_t b = 0;

    dataPin = (1 << dataPin);
    clockPin = (1 << clockPin);

    if (bitOrder == MSBFIRST) {
        for (i = 0; i < 8; i++) {
            OUTA |= clockPin;
            b <<= 1;
            if ((INA & dataPin) != 0) {
                b |= 1;
            }
            OUTA &= ~clockPin;
        }
    }
    else {
        for (i = 0; i < 8; i++) {
            OUTA |= clockPin;
            b >>= 1;
            if ((INA & dataPin) != 0) {
                b |= 0x80;
            }
            OUTA &= ~clockPin;
        }
    }

    return b;
}

void i2cStart()
{
    OUTA |= (1 << I2C_SCL);                        // Initially drive SCL HIGH
    DIRA |= (1 << I2C_SCL);
    OUTA |= (1 << I2C_SDA);                        // Initially drive SDA HIGH
    DIRA |= (1 << I2C_SDA);
    OUTA &= ~(1 << I2C_SDA);                       // Now drive SDA LOW
    OUTA &= ~(1 << I2C_SCL);                       // Leave SCL LOW
}

void i2cStop()
{
    OUTA |= (1 << I2C_SCL);                        // Drive SCL HIGH
    OUTA |= (1 << I2C_SDA);                        // then SDA HIGH
    DIRA &= ~(1 << I2C_SCL);                       // Now let them float
    DIRA &= ~(1 << I2C_SDA);                       // If pullups present, they'll stay HIGH
}

int i2cWrite(uint8_t data)
{
    int32_t i, ack = 0;

    for (i = 0x80; i != 0x00; i >>= 1) {    // Output data to SDA
        if ((data & i) != 0)
            OUTA |= (1 << I2C_SDA);
        else
            OUTA &= ~(1 << I2C_SDA);
        OUTA |= (1 << I2C_SCL);                        // Toggle SCL from LOW to HIGH to LOW
        __asm ("nop\r\n");
        __asm ("nop\r\n");
        OUTA &= ~(1 << I2C_SCL);
    }

    DIRA &= ~(1 << I2C_SDA);                           // Set SDA to input for ACK/NAK
    OUTA |= (1 << I2C_SCL);
    ack = INA & (1 << I2C_SDA);                        // Sample SDA when SCL is HIGH
    OUTA &= ~(1 << I2C_SCL);

    OUTA &= ~(1 << I2C_SDA);                           // Leave SDA driven LOW
    DIRA |= (1 << I2C_SDA);

    return ack == 0 ? I2C_ACK : I2C_NAK;
}

uint8_t i2cRead(int ack)
{
    int32_t i, data = 0;

    DIRA &= ~(1 << I2C_SDA);                       // Make SDA an input
    for (i = 0; i < 8; i++) {           // Receive data from SDA
        OUTA |= (1 << I2C_SCL);                    // Sample SDA when SCL is HIGH
        data <<= 1;
        data |= (INA & (1 << I2C_SDA)) >> I2C_SDA;
        OUTA &= ~(1 << I2C_SCL);
        __asm ("nop\r\n");
        __asm ("nop\r\n");
    }

    if (ack)                            // Output ACK/NAK to SDA
        OUTA |= (1 << I2C_SDA);
    else
        OUTA &= ~(1 << I2C_SDA);
    DIRA |= (1 << I2C_SDA);
    OUTA |= (1 << I2C_SCL);                        // Toggle SCL from LOW to HIGH to LOW
    OUTA &= ~(1 << I2C_SCL);

    OUTA &= ~(1 << I2C_SDA);                       // Leave SDA driven LOW

    return data;
}

void loadOverlay(int n)
{
    static int _ovly_current = -1;

    if (_ovly_current == n) {
        return;
    }

    uint32_t addr = HIGH_EEPROM_OFFSET(_ovly_table[n][0]);
    uint8_t * ptr = (uint8_t *)_ovly_addr;
    uint32_t count = _ovly_table[n][1];

    while (count > 0) {
        i2cStart();    // Select the device & send address
        i2cWrite(EEPROM_ADDR | I2C_WRITE | ((addr & 0x10000) >> 15));
        i2cWrite(addr >> 8);
        i2cWrite(addr & 0xFF);

        i2cStart();    // Reselect the device for reading
        i2cWrite(EEPROM_ADDR | I2C_READ | ((addr & 0x10000) >> 15));

        addr++;
        while (count > 1 && (addr & 0xFFFF) != 0) {
            *ptr++ = i2cRead(I2C_ACK);
            count--;
            addr++;
        }
        *ptr++ = i2cRead(I2C_NAK);
        count--;

        i2cStop();
    }

    _ovly_current = n;
}
