/*
 * Transmit / Receive Size limits.
 *
 * Transmit size limit is based on free Cog RAM. It can be increased if we save space
 * in the cog by optimizing the code or removing other data. Receive size is limited only
 * by available hub ram.
 *
 * Note that if TX_BUFFER_WORDS is too large the error is detected at compile-time, but
 * if RX_BUFFER_WORDS is too large we won't detect the error until Start is running!
 */

#define TX_BUFFER_WORDS     185 // FIXME is this right?
#define RX_BUFFER_WORDS     266

/*
 * Maximum stored configuration descriptor size, in bytes. If the configuration
 * descriptor is longer than this, we'll truncate it. Must be a multiple of 4.
 */

#define CFGDESC_BUFFER_LEN  256

/*
 * USB data pins.
 *
 * Important: Both DMINUS and DPLUS must be <= 7, since we
 *            use pin masks in instruction literals, and we
 *            assume we're using the first video generator bank.
 */

#define DMINUS  1
#define DPLUS   0

// This no-connect pin is used for cog to cog and counter to counter signalling.
#define USBNC   2

/*
 * This module can be very challenging to debug. To make things a little bit easier,
 * there are several places where debug pin masks are OR'ed into our DIRA values
 * at compile-time. This doesn't use any additional memory. With a logic analyzer,
 * you can use this to see exactly when the bus is being driven, and by what code.
 *
 * To use this, pick some pin(s) to use, put their bit masks here. Attach a pull-up
 * resistor and logic analyzer probe to each pin. To disable, set all values to zero.
 *
 * Since the masks must fit in an instruction's immediate data, you must use P0 through P8.
 */

#define DEBUG_ACK_MASK   0
#define DEBUG_TX_MASK    0

// Low-level debug flags, settable at runtime

#define DEBUGFLAG_NO_CRC 0x01

// Output bus states

#define BUS_MASK    ((1 << DPLUS) | (1 << DMINUS))
#define STATE_J     (1 << DPLUS)
#define STATE_K     (1 << DMINUS)
#define STATE_SE0   0
#define STATE_SE1   BUS_MASK

// Retry options

#define MAX_TOKEN_RETRIES      200
#define MAX_CRC_RETRIES        200
#define TIMEOUT_FRAME_DELAY    10

// Number of CRC error retry attempts

// Offsets in EndpointTable

#define EPTABLE_SHIFT        2        // log2 of entry size
#define EPTABLE_TOKEN        0        // word
#define EPTABLE_TOGGLE_IN    2        // byte
#define EPTABLE_TOGGLE_OUT   3        // byte

// Port connection status codes

#define PORTC_NO_DEVICE    STATE_SE0     // No device (pull-down resistors in host)
#define PORTC_FULL_SPEED   STATE_J       // Full speed: pull-up on D+
#define PORTC_LOW_SPEED    STATE_K       // Low speed: pull-up on D-
#define PORTC_INVALID      STATE_SE1     // Buggy device? Electrical transient?
#define PORTC_NOT_READY    0xFF          // Haven't checked port status yet

// Command opcodes for the controller cog.

#define OP_NOP             0           // Do nothing
#define OP_RESET           1           // Send a USB Reset signal   '
#define OP_TX_BEGIN        2           // Start a TX packet. Includes 8-bit PID
#define OP_TX_END          3           // End a TX packet, arg = # of extra idle bits after
#define OP_TXRX            4           // Transmit and/or receive packets
#define OP_TX_DATA_16      5           // Encode and store a 16-bit word
#define OP_TX_DATA_PTR     6           // Encode data from hub memory.
                                       //   Command arg: pointer
                                       //   "result" IN: Number of bytes
#define OP_TX_CRC16        7           // Encode  a 16-bit CRC of all data since the PID
#define OP_RX_PID          8           // Decode and return a 16-bit PID word, reset CRC-16
#define OP_RX_DATA_PTR     9           // Decode data to hub memory.
                                       //   Command arg: pointer
                                       //   "result" IN: Max number of bytes
                                       //   result OUT:  Actual number of bytes
#define OP_RX_CRC16        10          // Decode and check CRC. Returns (actual XOR expected)
#define OP_SOF_WAIT        11          // Wait for one SOF to be sent

// OP_TXRX parameters

#define TXRX_TX_ONLY       0b00
#define TXRX_TX_RX         0b01
#define TXRX_TX_RX_ACK     0b11

// USB PID values / commands

#define PID_OUT      0b11100001
#define PID_IN       0b01101001
#define PID_SOF      0b10100101
#define PID_SETUP    0b00101101
#define PID_DATA0    0b11000011
#define PID_DATA1    0b01001011
#define PID_ACK      0b11010010
#define PID_NAK      0b01011010
#define PID_STALL    0b00011110
#define PID_PRE      0b00111100

/*
 * NRZI-decoded representation of a SYNC field, and PIDs which include the SYNC.
 * These are the form of values returned by OP_RX_PID.
 */

#define SYNC_FIELD        0b10000000
#define SYNC_PID_ACK      SYNC_FIELD | (PID_ACK << 8)
#define SYNC_PID_NAK      SYNC_FIELD | (PID_NAK << 8)
#define SYNC_PID_STALL    SYNC_FIELD | (PID_STALL << 8)
#define SYNC_PID_DATA0    SYNC_FIELD | (PID_DATA0 << 8)
#define SYNC_PID_DATA1    SYNC_FIELD | (PID_DATA1 << 8)

/*
 * USB Tokens (Device ID + Endpoint) with pre-calculated CRC5 values.
 * Since we only support a single USB device, we only need tokens for
 * device 0 (the default address) and device 1 (our arbitrary device ID).
 * For device 0, we only need endpoint zero. For device 1, we include
 * tokens for every possible endpoint.
 */

//                          CRC5 EP# DEV#
#define TOKEN_DEV0_EP0    0b0001000000000000
#define TOKEN_DEV1_EP0    0b1110100000000001
#define TOKEN_DEV1_EP1    0b0101100010000001
#define TOKEN_DEV1_EP2    0b1100000100000001
#define TOKEN_DEV1_EP3    0b0111000110000001
#define TOKEN_DEV1_EP4    0b1011101000000001
#define TOKEN_DEV1_EP5    0b0000101010000001
#define TOKEN_DEV1_EP6    0b1001001100000001
#define TOKEN_DEV1_EP7    0b0010001110000001
#define TOKEN_DEV1_EP8    0b0100110000000001
#define TOKEN_DEV1_EP9    0b1111110010000001
#define TOKEN_DEV1_EP10   0b0110010100000001
#define TOKEN_DEV1_EP11   0b1101010110000001
#define TOKEN_DEV1_EP12   0b0001111000000001
#define TOKEN_DEV1_EP13   0b1010111010000001
#define TOKEN_DEV1_EP14   0b0011011100000001
#define TOKEN_DEV1_EP15   0b1000011110000001

/*
 * Standard device requests.
 *
 * This encodes the first two bytes of the SETUP packet into
 * one word-sized command. The low byte is bmRequestType,
 * the high byte is bRequest.
 */

#define REQ_CLEAR_DEVICE_FEATURE       0x0100
#define REQ_CLEAR_INTERFACE_FEATURE    0x0101
#define REQ_CLEAR_ENDPOINT_FEATURE     0x0102
#define REQ_GET_CONFIGURATION          0x0880
#define REQ_GET_DESCRIPTOR             0x0680
#define REQ_GET_INTERFACE              0x0a81
#define REQ_GET_DEVICE_STATUS          0x0000
#define REQ_GET_INTERFACE_STATUS       0x0001
#define REQ_GET_ENDPOINT_STATUS        0x0002
#define REQ_SET_ADDRESS                0x0500
#define REQ_SET_CONFIGURATION          0x0900
#define REQ_SET_DESCRIPTOR             0x0700
#define REQ_SET_DEVICE_FEATURE         0x0300
#define REQ_SET_INTERFACE_FEATURE      0x0301
#define REQ_SET_ENDPOINT_FEATURE       0x0302
#define REQ_SET_INTERFACE              0x0b01
#define REQ_SYNCH_FRAME                0x0c82

/*
 * Standard descriptor types.
 *
 * These identify a descriptor in REQ_GET_DESCRIPTOR,
 * via the high byte of wValue. (wIndex is the language ID.)
 *
 * The 'DESCHDR' variants are the full descriptor header,
 * including type and length. This matches the first two bytes
 * of any such static-length descriptor.
 */

#define DESC_DEVICE             0x0100
#define DESC_CONFIGURATION      0x0200
#define DESC_STRING             0x0300
#define DESC_INTERFACE          0x0400
#define DESC_ENDPOINT           0x0500

#define DESCHDR_DEVICE          0x0112
#define DESCHDR_CONFIGURATION   0x0209
#define DESCHDR_INTERFACE       0x0409
#define DESCHDR_ENDPOINT        0x0507

// Descriptor Formats

#define DEVDESC_bLength               0
#define DEVDESC_bDescriptorType       1
#define DEVDESC_bcdUSB                2
#define DEVDESC_bDeviceClass          4
#define DEVDESC_bDeviceSubClass       5
#define DEVDESC_bDeviceProtocol       6
#define DEVDESC_bMaxPacketSize0       7
#define DEVDESC_idVendor              8
#define DEVDESC_idProduct             10
#define DEVDESC_bcdDevice             12
#define DEVDESC_iManufacturer         14
#define DEVDESC_iProduct              15
#define DEVDESC_iSerialNumber         16
#define DEVDESC_bNumConfigurations    17
#define DEVDESC_LEN                   18

#define CFGDESC_bLength               0
#define CFGDESC_bDescriptorType       1
#define CFGDESC_wTotalLength          2
#define CFGDESC_bNumInterfaces        4
#define CFGDESC_bConfigurationValue   5
#define CFGDESC_iConfiguration        6
#define CFGDESC_bmAttributes          7
#define CFGDESC_MaxPower              8

#define IFDESC_bLength                0
#define IFDESC_bDescriptorType        1
#define IFDESC_bInterfaceNumber       2
#define IFDESC_bAlternateSetting      3
#define IFDESC_bNumEndpoints          4
#define IFDESC_bInterfaceClass        5
#define IFDESC_bInterfaceSubClass     6
#define IFDESC_bInterfaceProtocol     7
#define IFDESC_iInterface             8

#define EPDESC_bLength                0
#define EPDESC_bDescriptorType        1
#define EPDESC_bEndpointAddress       2
#define EPDESC_bmAttributes           3
#define EPDESC_wMaxPacketSize         4
#define EPDESC_bInterval              6

// SETUP packet format

#define SETUP_bmRequestType           0
#define SETUP_bRequest                1
#define SETUP_wValue                  2
#define SETUP_wIndex                  4
#define SETUP_wLength                 6
#define SETUP_LEN                     8

// Endpoint constants

#define DIR_IN         0x80
#define DIR_OUT        0x00

#define TT_CONTROL     0x00
#define TT_ISOC        0x01
#define TT_BULK        0x02
#define TT_INTERRUPT   0x03

/*
 * Negative error codes. Most functions in this library can call
 * "abort" with one of these codes.
 *
 * So that multiple levels of the software stack can share
 * error codes, we define a few reserved ranges:
 *
 *   -1 to -99    : Application
 *   -100 to -150 : Device or device class driver
 *   -150 to -199 : USB host controller
 *
 * Within the USB host controller range:
 *
 *   -150 to -159 : Device connectivity errors
 *   -160 to -179 : Low-level transfer errors
 *   -180 to -199 : High-level errors (parsing, resource exhaustion)
 *
 * When adding new errors, please keep existing errors constant
 * to avoid breaking other modules who may depend on these values.
 * (But if you're writing another module that uses these values,
 * please use the constants from this object rather than hardcoding
 * them!)
 */

#define E_SUCCESS         0

#define E_NO_DEVICE       -150        // No device is attached
#define E_LOW_SPEED       -151        // Low-speed devices are not supported
#define E_PORT_BOUNCE     -152        // Port connection state changing during Enumerate

#define E_TIMEOUT         -160        // Timed out waiting for a response
#define E_TRANSFER        -161        // Generic low-level transfer error
#define E_CRC             -162        // CRC-16 mismatch and/or babble condition
#define E_TOGGLE          -163        // DATA0/1 toggle error
#define E_PID             -164        // Invalid or malformed PID and/or no response
#define E_STALL           -165        // USB STALL response (pipe error)

#define E_DEV_ADDRESS     -170        // Enumeration error: Device addressing
#define E_READ_DD_1       -171        // Enumeration error: First device descriptor read
#define E_READ_DD_2       -172        // Enumeration error: Second device descriptor read
#define E_READ_CONFIG     -173        // Enumeration error: Config descriptor read

#define E_OUT_OF_COGS     -180        // Not enough free cogs, can't initialize
#define E_OUT_OF_MEM      -181        // Not enough space for the requested buffer sizes
#define E_DESC_PARSE      -182        // Can't parse a USB descriptor
