/*
 * This is a software implementation of a simple full-speed
 * (12 Mb/s) USB 1.1 host controller for the Parallax Propeller.
 *
 * Based on original code by M. Elizabeth Scott
 * https://github.com/scanlime/propeller-usb-host
 *
 * Ported to C / C++ by Marco Maccaferri.
 *
 * Copyright (c) 2010-2016 M. Elizabeth Scott <beth@scanlime.org>
 * Copyright (c) 2020 M. Maccaferri <macca@maccasoft.com>
 *
 * See end of file for terms of use.
 */

#include <propeller.h>

#include "USBHost.h"

#ifdef __GNUC__
#define INLINE__ static inline
#define Yield__() __asm__ volatile( "" ::: "memory" )
#else
#define INLINE__ static
static int32_t tmp__;
#define Yield__()
#endif

INLINE__ int32_t Rotl__(uint32_t a, uint32_t b) {
    return (a << b) | (a >> (32 - b));
}
INLINE__ int32_t Rotr__(uint32_t a, uint32_t b) {
    return (a >> b) | (a << (32 - b));
}
INLINE__ int32_t Lookup__(int32_t x, int32_t b, int32_t a[], int32_t n) {
    int32_t i = (x) - (b);
    return ((unsigned) i >= n) ? 0 : (a)[i];
}

INLINE__ uint16_t UWORD(uint8_t * ptr) {
    return ptr[0] | (ptr[1] << 8);
}

INLINE__ void UWORD(uint8_t * ptr, uint16_t val) {
    ptr[0] = val;
    ptr[1] = val >> 8;
}

INLINE__ void UDWORD(uint8_t * ptr, uint32_t val) {
    ptr[0] = val;
    ptr[1] = val >> 8;
    ptr[2] = val >> 16;
    ptr[3] = val >> 24;
}

struct _ep {
    uint16_t token;
    uint8_t  in_toggle;
    uint8_t  out_toggle;
};

struct _ep EndpointTable[] = {
    { TOKEN_DEV1_EP0, 0, 0 },
    { TOKEN_DEV1_EP1, 0, 0 },
    { TOKEN_DEV1_EP2, 0, 0 },
    { TOKEN_DEV1_EP3, 0, 0 },
    { TOKEN_DEV1_EP4, 0, 0 },
    { TOKEN_DEV1_EP5, 0, 0 },
    { TOKEN_DEV1_EP6, 0, 0 },
    { TOKEN_DEV1_EP7, 0, 0 },
    { TOKEN_DEV1_EP8, 0, 0 },
    { TOKEN_DEV1_EP9, 0, 0 },
    { TOKEN_DEV1_EP10, 0, 0 },
    { TOKEN_DEV1_EP11, 0, 0 },
    { TOKEN_DEV1_EP12, 0, 0 },
    { TOKEN_DEV1_EP13, 0, 0 },
    { TOKEN_DEV1_EP14, 0, 0 },
    { TOKEN_DEV1_EP15, 0, 0 }
};

USBHost::USBHost()
{

}

/* ============================================================================== *
 * Host Controller Setup                                                          *
 * ============================================================================== */

int USBHost::start()
{
    txc_command = -1;
    rx1_time = -1;
    rx2_sop = -1;
    portc = PORTC_NOT_READY;
    rxdone = 0xFF;

    alloc_buf_dd = (uint8_t *) _load_start_cog_usb_host_fs_controller;
    alloc_buf_cfg = alloc_buf_dd + DEVDESC_LEN;
    alloc_buf_setup = alloc_buf_dd + CFGDESC_BUFFER_LEN;
    alloc_buf_rtx = alloc_buf_setup + SETUP_LEN;

    buf_dd = (uint16_t)(uint32_t)alloc_buf_dd;
    buf_cfg = (uint16_t)(uint32_t)alloc_buf_cfg;
    buf_setup = (uint16_t)(uint32_t)alloc_buf_setup;

    _load_start_cog_usb_host_fs_controller[(uint32_t)&sof_deadline >> 2] = CNT;
    _load_start_cog_usb_host_fs_controller[(uint32_t)&txp_portc >> 2] = (uint32_t)&portc;
    _load_start_cog_usb_host_fs_controller[(uint32_t)&txp_result >> 2] = (uint32_t)&txc_result;
    _load_start_cog_usb_host_fs_controller[(uint32_t)&txp_rxdone >> 2] = (uint32_t)&rxdone;
    _load_start_cog_usb_host_fs_controller[(uint32_t)&txp_rx1_time >> 2] = (uint32_t)&rx1_time;
    _load_start_cog_usb_host_fs_controller[(uint32_t)&txp_rxbuffer >> 2] = (uint32_t)alloc_buf_rtx;

    _load_start_cog_usb_host_fs_receiver[(uint32_t)&rx1p_done >> 2] = (uint32_t)&rxdone;
    _load_start_cog_usb_host_fs_receiver[(uint32_t)&rx1p_buffer >> 2] = (uint32_t)alloc_buf_rtx;
    _load_start_cog_usb_host_fs_receiver[(uint32_t)&rx1p_sop >> 2] = (uint32_t)&rx2_sop;
    _load_start_cog_usb_host_fs_receiver[(uint32_t)&rx1p_portc >> 2] = (uint32_t)&portc;

    cog1 = cognew(_load_start_cog_usb_host_fs_controller, (uint32_t)&txc_command);
    cog2 = cognew(_load_start_cog_usb_host_fs_receiver, (uint32_t)&rx1_time);

    while(txc_result || rx1_time) {
        Yield__();
    }

    return 1;
}

/*
 * Wait for the controller to send 'count' Start Of Frame tokens.
 * If one SOF has been emitted since the last call to frame_wait, it may
 * count as the first in 'count'.
 */
void USBHost::frame_wait(int count)
{
    while (count > 0) {
        command(OP_SOF_WAIT, 0);
        count--;
    }
    sync();
}

/* ==============================================================================
 * High-level Device Framework
 * ============================================================================== */

/*
 * Is a device connected? If so, what speed? Returns a PORTC_* constant.
 */
int USBHost::get_port_connection()
{
    while (portc == PORTC_NOT_READY) {
        Yield__();
    }
    return portc;
}

/*
 * Initialize the attached USB device, and get information about it.
 *
 *   1. Reset the device
 *   2. Assign it an address
 *   3. Read the device descriptor
 *   4. Read the first configuration descriptor
 */
int USBHost::enumerate()
{
    int rc;

    // Port debounce: Make sure the device is in the
    // same connection state for a couple frames.
    int pc = get_port_connection();
    frame_wait(3);
    if (get_port_connection() != pc) {
        return E_PORT_BOUNCE;
    }

    if (pc == PORTC_NO_DEVICE || pc == PORTC_INVALID) {
        return E_NO_DEVICE;
    }

    // Device reset, and give it some time to wake up
    device_reset();
    frame_wait(10);

    /*
     * Before we can do any transfers longer than 8 bytes, we need to know the maximum
     * packet size on EP0. Otherwise we won't be able to determine when a transfer has
     * ended. So, we'll use a temporary maximum packet size of 8 in order to address the
     * device and to receive the first 8 bytes of the device descriptor. This should
     * always be possible using transfers of no more than one packet in length.
     */
    alloc_buf_dd[DEVDESC_bMaxPacketSize0] = 8;

    rc = device_address();
    if (0 > rc) {
        return E_DEV_ADDRESS;
    }

    // Read the real max packet length (Must request exactly 8 bytes)
    if (control_read(REQ_GET_DESCRIPTOR, DESC_DEVICE, 0, alloc_buf_dd, 8) < 0) {
        return E_READ_DD_1;
    }

    // Validate device descriptor header
    if (UWORD(&alloc_buf_dd[0]) != DESCHDR_DEVICE) {
        return E_DESC_PARSE;
    }

    // Read the whole descriptor
    if (control_read(REQ_GET_DESCRIPTOR, DESC_DEVICE, 0, alloc_buf_dd, DEVDESC_LEN) < 0) {
        return E_READ_DD_2;
    }

    read_configuration(0);

    return E_SUCCESS;
}

int USBHost::configure()
{
    reset_endpoint_toggle();
    return control(REQ_SET_CONFIGURATION, alloc_buf_cfg[CFGDESC_bConfigurationValue], 0);
}

int USBHost::unconfigure()
{
    return control(REQ_SET_CONFIGURATION, 0, 0);
}

int USBHost::read_configuration(int index)
{
    return control_read(REQ_GET_DESCRIPTOR, DESC_CONFIGURATION | index, 0, alloc_buf_cfg, CFGDESC_BUFFER_LEN);
}

int USBHost::vendor_id()
{
    return UWORD(&alloc_buf_dd[DEVDESC_idVendor]);
}

int USBHost::product_id()
{
    return UWORD(&alloc_buf_dd[DEVDESC_idProduct]);
}

/* Clear a Halt condition on one endpoint, given a pointer to the endpoint descriptor */

void USBHost::clear_halt(uint8_t * epd)
{
    control(REQ_CLEAR_ENDPOINT_FEATURE, 0, epd[EPDESC_bEndpointAddress]);
}

/* ============================================================================== *
 * Configuration Descriptor Parsing                                               *
 * ============================================================================== */

uint8_t * USBHost::next_descriptor(uint8_t * ptrIn)
{
    int length = UWORD(&alloc_buf_cfg[CFGDESC_wTotalLength]);
    if (length > CFGDESC_BUFFER_LEN) {
        length = CFGDESC_BUFFER_LEN;
    }
    uint8_t * endPtr = alloc_buf_cfg + length;

    uint8_t * ptrOut = ptrIn + ptrIn[0];
    if (ptrOut >= endPtr) {
        ptrOut = NULL;
    }

    return ptrOut;
}

uint8_t * USBHost::next_header_match(uint8_t * ptrIn, uint16_t header)
{
    while ((ptrIn = next_descriptor(ptrIn)) != NULL) {
        if (UWORD(ptrIn) == header)
            return ptrIn;
    }
    return NULL;
}

uint8_t * USBHost::first_interface()
{
    return next_interface(alloc_buf_cfg);
}

uint8_t * USBHost::next_interface(uint8_t * curIf)
{
    return next_header_match(curIf, DESCHDR_INTERFACE);
}

uint8_t * USBHost::next_endpoint(uint8_t * curIf)
{
    while ((curIf = next_descriptor(curIf)) != NULL) {
        switch(UWORD(curIf)) {
            case DESCHDR_ENDPOINT:
                return curIf;
            case DESCHDR_INTERFACE:
                return NULL;
        }
    }
    return NULL;
}

uint8_t * USBHost::find_interface(int ifClass)
{
    uint8_t * foundIf = first_interface();
    while (foundIf != NULL) {
        if (foundIf[IFDESC_bInterfaceClass] == ifClass) {
            return foundIf;
        }
        foundIf = next_interface(foundIf);
    }
    return NULL;
}

/*

PUB EndpointDirection(epd)
  '' Given an endpoint descriptor pointer, test the endpoint direction.
  '' (DIR_IN or DIR_OUT)

  return BYTE[epd + EPDESC_bEndpointAddress] & $80

PUB EndpointType(epd)
  '' Return an endpoint's transfer type (TT_BULK, TT_ISOC, TT_INTERRUPT)

  return BYTE[epd + EPDESC_bmAttributes] & $03


DAT
*/

/* ==============================================================================
 * Device Setup
 * ============================================================================== */

/*
 * Asynchronously send a USB bus reset signal.
 */
void USBHost::device_reset()
{
    command(OP_RESET, 0);
    reset_endpoint_toggle();
}

/*
 * Send a SET_ADDRESS(1) to device 0.
 *
 * This should be sent after device_reset to transition the
 * device from the Default state to the Addressed state. All
 * other transfers here assume the device address is 1.
 */
int USBHost::device_address()
{
    uint32_t buf;

    UWORD(&alloc_buf_setup[0], REQ_SET_ADDRESS);
    UWORD(&alloc_buf_setup[SETUP_wValue], 1);
    UDWORD(&alloc_buf_setup[SETUP_wIndex], 0);

    return control_raw(TOKEN_DEV0_EP0, (uint8_t *)&buf, 4);
}

/* ==============================================================================
 * Control Transfers
 * ============================================================================== */

/*
 * Issue a no-data control transfer to an addressed device.
 */
int USBHost::control(int req, int value, int index)
{
    uint32_t buf;

    UWORD(&alloc_buf_setup[0], req);
    UWORD(&alloc_buf_setup[SETUP_wValue], value);
    UWORD(&alloc_buf_setup[SETUP_wIndex], index);
    UWORD(&alloc_buf_setup[SETUP_wLength], 0);

    return control_raw(TOKEN_DEV1_EP0, (uint8_t *)&buf, 4);
}

/* Issue a control IN transfer to an addressed device.
 *
 * Returns the number of bytes read.
 * Aborts on error.
 */
int USBHost::control_read(int req, int value, int index, uint8_t *bufferPtr, int length)
{
    int toggle, result;

    UWORD(&alloc_buf_setup[0], req);
    UWORD(&alloc_buf_setup[SETUP_wValue], value);
    UWORD(&alloc_buf_setup[SETUP_wIndex], index);
    UWORD(&alloc_buf_setup[SETUP_wLength], length);

    // Issues SETUP and IN transactions
    result = control_raw(TOKEN_DEV1_EP0, bufferPtr, length);

    // Status phase (OUT + DATA1)
    toggle = PID_DATA1;
    write_data(PID_OUT, TOKEN_DEV1_EP0, 0, 0, (uint8_t *)&toggle, MAX_TOKEN_RETRIES);

    return result;
}

/*
 * Issue a control OUT transfer to an addressed device.
 */
int USBHost::control_write(int req, int value, int index, uint8_t *bufferPtr, int length)
{
    int toggle, pktSize0, packetSize;

    UWORD(&alloc_buf_setup[0], req);
    UWORD(&alloc_buf_setup[SETUP_wValue], value);
    UWORD(&alloc_buf_setup[SETUP_wIndex], index);
    UWORD(&alloc_buf_setup[SETUP_wLength], length);

    toggle = PID_DATA0;
    write_data(PID_SETUP, TOKEN_DEV1_EP0, alloc_buf_setup, 8, (uint8_t*) &toggle, MAX_TOKEN_RETRIES);

    // Break OUT data into multiple packets if necessary
    pktSize0 = alloc_buf_dd[DEVDESC_bMaxPacketSize0];
    while (1) {
        packetSize = length;
        if (packetSize > pktSize0)
            packetSize = pktSize0;
        write_data(PID_OUT, TOKEN_DEV1_EP0, bufferPtr, packetSize, (uint8_t*) &toggle, MAX_TOKEN_RETRIES);
        bufferPtr += packetSize;
        if ((length -= packetSize) <= 0) {

            // Status stage (always DATA1)
            toggle = PID_DATA1;
            return data_in(TOKEN_DEV1_EP0, (uint8_t*) &packetSize, 4, pktSize0, (uint8_t*) &toggle, TXRX_TX_RX_ACK, MAX_TOKEN_RETRIES, 1);
        }
    }
}

/*
 * Common low-level implementation of no-data and read control transfers.
 */
int USBHost::control_raw(int token, uint8_t *buffer, int length)
{
    uint32_t toggle = PID_DATA0;
    write_data(PID_SETUP, token, alloc_buf_setup, 8, (uint8_t *)&toggle, MAX_TOKEN_RETRIES);
    return data_in(token, buffer, length, alloc_buf_dd[DEVDESC_bMaxPacketSize0], (uint8_t *)&toggle, TXRX_TX_RX_ACK, MAX_TOKEN_RETRIES, 1);
}

/* ============================================================================== *
 * Interrupt Transfers                                                            *
 * ============================================================================== */

/*
 * Try to read one packet, up to 'length' bytes, from an Interrupt IN endpoint.
 * Returns the actual amount of data read.
 * If no data is available, raises E_TIMEOUT without waiting.
 *
 * 'epd' is a pointer to this endpoint's Endpoint Descriptor.
 *
 * This is different from Bulk in two main ways:
 *
 *   - We give data_in an artificially large maxPacketSize, since we
 *     never want it to receive more than one packet at a time here.
 *   - We give it a retry of 0, since we don't want to retry on NAK.
 */
int USBHost::interrupt_read(uint8_t * epd, uint8_t * buffer, int length)
{
    struct _ep * epTable = &EndpointTable[epd[EPDESC_bEndpointAddress] & 0x0F];
    return data_in(epTable->token, buffer, length, 0x1000, (uint8_t *)&epTable->in_toggle, TXRX_TX_RX, 0, MAX_CRC_RETRIES);
}

/* ============================================================================== *
 * Bulk Transfers                                                                 *
 * ============================================================================== */

/*
 * Write 'length' bytes of data to a Bulk OUT endpoint.
 *
 * Always writes at least one packet. If 'length' is zero,
 * we send a zero-length packet. If 'length' is any other
 * even multiple of maxPacketLen, we send only maximally-long
 * packets and no zero-length packet.
 *
 * 'epd' is a pointer to this endpoint's Endpoint Descriptor.
 */
void USBHost::bulk_write(uint8_t * epd, uint8_t * buffer, int length)
{
    struct _ep * ep_table = &EndpointTable[epd[EPDESC_bEndpointAddress] & 0x0F];
    int max_packet_size = UWORD(epd + EPDESC_wMaxPacketSize);

    while (1) {
        int packet_size = length;
        if (length > max_packet_size) {
            length = max_packet_size;
        }
        write_data(PID_OUT, (int)ep_table, buffer, packet_size, (uint8_t *)&ep_table->out_toggle, MAX_TOKEN_RETRIES);

        buffer += packet_size;
        if ((length -= packet_size) <= 0) {
            return;
        }
    }
}

/*
 * Read up to 'length' bytes from a Bulk IN endpoint.
 * Returns the actual amount of data read.
 *
 * 'epd' is a pointer to this endpoint's Endpoint Descriptor.
 */
int USBHost::bulk_read(uint8_t * epd, uint8_t * buffer, int length)
{
    struct _ep * ep_table = &EndpointTable[epd[EPDESC_bEndpointAddress] & 0x0F];
    return data_in(ep_table->token, buffer, length, UWORD(epd + EPDESC_wMaxPacketSize), (uint8_t *)&ep_table->in_toggle, TXRX_TX_RX, MAX_TOKEN_RETRIES, MAX_CRC_RETRIES);
}

/* ============================================================================== *
 * Low-level Transfer Utilities                                                   *
 * ============================================================================== */

/*
 * Reset all endpoints to the default DATA0 toggle
 */
void USBHost::reset_endpoint_toggle()
{
    for (int ep = 0; ep < 16; ep++) {
        EndpointTable[ep].in_toggle = EndpointTable[ep].out_toggle = PID_DATA0;
    }
}

/*
 * Issue IN tokens and read the resulting data packets until
 * a packet smaller than maxPacketLen arrives. On success,
 * returns the actual number of bytes read. On failure, returns
 * a negative error code.
 *
 * 'togglePtr' is a pointer to a byte with either PID_DATA0 or
 * PID_DATA1, depending on which DATA PID we expect next. Every
 * time we receive a packet, we toggle this byte from DATA0 to
 * DATA1 or vice versa.
 *
 * Each packet will have up to 'retries' additional attempts
 * if the device responds with a NAK.
 */

int USBHost::data_in(int token, uint8_t *buffer, int length, int maxPacketLen, uint8_t *togglePtr, int txrxFlag, int tokenRetries, int crcRetries)
{
    int actual, packetLen;

    actual = 0;

    /*
     * As long as there's buffer space, send IN tokens. Each IN token
     * allows the device to send us back up to maxPacketLen bytes of data.
     * If the device sends a short packet (including zero-byte packets)
     * it terminates the transfer.
     */

    while (1) {
        packetLen = read_data_in(token, buffer, length, togglePtr, txrxFlag, tokenRetries, crcRetries);
        actual += packetLen;
        buffer += packetLen;
        length -= packetLen;

        if (packetLen < maxPacketLen) {
            return actual;
            //return -1; // Short packet. Device ended the transfer early.
        }
        if (length <= 0) {
            return actual;
            //return E_SUCCESS; // Transfer fully completed
        }
    }
}

/*
 * Transmit a single data packet to an endpoint, as a token followed by DATA.
 *
 * 'togglePtr' is a pointer to a byte with either PID_DATA0 or
 * PID_DATA1, depending on which DATA PID we expect next. Every
 * time we receive a packet, we toggle this byte from DATA0 to
 * DATA1 or vice versa.
 *
 * Each packet will have up to 'retries' additional attempts
 * if the device responds with a NAK.
 */

int USBHost::write_data(int pid, int token, uint8_t *buffer, int length, uint8_t *togglePtr, int retries)
{
    while (1) {
        send_token(pid, token, 10);
        command(OP_TX_BEGIN, *togglePtr);      // DATA0/1

        if (length) {
            sync();
            txc_result = length;
            command(OP_TX_DATA_PTR, (int) buffer);
        }

        command(OP_TX_CRC16, 0);
        command(OP_TX_END, 1);
        command(OP_TXRX, TXRX_TX_RX);

        command(OP_RX_PID, 0);
        sync();
        switch (txc_result) {

            case SYNC_PID_NAK:
                // Busy. Wait a frame and try again.
                if (--retries <= 0) {
                    return E_TIMEOUT;
                }
                frame_wait(TIMEOUT_FRAME_DELAY);
                break;

            case SYNC_PID_STALL:
                return E_STALL;

            case SYNC_PID_ACK:
                *togglePtr ^= (PID_DATA0 ^ PID_DATA1);
                return E_SUCCESS;

            default:
                last_pid_err = txc_result;
                return E_PID;
        }
    }
}

/*
 * Low-level data IN request. Handles data toggle and retry.
 * This is part of the implementation of data_in().
 * Aborts on error, otherwise returns the EOP timestamp.
 */
int USBHost::request_data_in(int token, int txrxFlag, uint8_t *togglePtr, int retries)
{
    int result = E_SUCCESS;

    while (1) {
        send_token(PID_IN, token, 0);

        command(OP_TXRX, txrxFlag);
        sync();
        result = txc_result;

        command(OP_RX_PID, 0);
        sync();
        switch (txc_result) {
            case SYNC_PID_NAK:
                // Busy. Wait a frame and try again.
                if (--retries <= 0) {
                    return E_TIMEOUT;
                }
                frame_wait(TIMEOUT_FRAME_DELAY);
                break;

            case SYNC_PID_STALL:
                return E_STALL;

            case SYNC_PID_DATA0:
            case SYNC_PID_DATA1:
                if ((txc_result >> 8) != *togglePtr) {
                    return E_TOGGLE;
                }
                if (txrxFlag == TXRX_TX_RX_ACK)
                    // Only toggle if we're ACK'ing this packet
                    *togglePtr ^= (PID_DATA0 ^ PID_DATA1);
                return result;

            default:
                last_pid_err = txc_result;
                return E_PID;
        }
    }
}

/*
 * Low-level data IN request + read to buffer.
 * This is part of the implementation of DataIN().
 * Aborts on error, otherwise returns the actual number of bytes read.
 *
 * This implements our crazy "deferred ACK" scheme, to work around
 * the fact that we can't decode a packet fast enough to check it prior
 * to sending the ACK when we're supposed to. In this strategy, we send
 * multiple IN tokens for each actual packet we intend to read, and postpone
 * the ACK by one packet. The first packet is never acknowledged. The second
 * will be ACK'ed only if the first was checked successfully. And so on.
 * Normally this means that there are two packets: The first one is decoded
 * but not ACKed, and the second one is ACKed but the contents are ignored.
 * (We assume it's identical to the first.)
 *
 * To use deferred ACKs, set txrxFlag to TXRX_TX_RX. To ACK the first
 * response (before the CRC check) set txrxFlag to TXRX_TX_RX_ACK.
 * If deferred ACKs are NOT in use, crcRetries must be 1.
 *
 * We're currently using deferred ACKs only for Bulk and Interrupt INs.
 * Some devices don't handle non-acknowledged IN packets during control
 * transfers, so we still have the possibility of CRC errors during control
 * reads. Luckily, most such transfers can be retried at a higher
 * protocol level.
 */

int USBHost::read_data_in(int token, uint8_t *buffer, int length, uint8_t *togglePtr, int txrxFlag, int tokenRetries, int crcRetries)
{
    int pc, result;

    while (crcRetries-- > 0) {
        /* The process of actually decoding the received packet is a bit convoluted,
         * due to the split of responsibilities between the RX cogs, TX cog, and
         * Spin code:
         *
         *   - The RX cogs don't really detect the EOP condition, they just stop when
         *     the D- line has been zero for a while. But they do record an accurate
         *     SOP timestamp.
         *
         *   - The TX cog records a fairly accurate EOP timestamp
         *
         *   - Spin code uses these timestamps to figure out the maximum number of
         *     bits that reside in a packet, minus the CRC and header.
         *
         *   - The decoder is given both a bit and byte limit, so it stops writing
         *     when either the buffer fills up or we reach the calculated EOP minus
         *     CRC position.
         *
         * Calculate the actual packet length.
         *
         * request_data_in returns an EOP timestamp. The TX cog knew when the EOP
         * was, and the RX2 cog knew when SOP was. We're running at 8 clocks
         * per bit, or 64 clocks per byte. Convert clocks to bits.
         *
         * We subtract enough bits to cover the CRC-16 and PID. The decoder needs us
         * to account for slightly less bits than we're actually receiving, since it
         * stops after the bit count underflows. To round evenly we could subtract half
         * a byte, but we only subtract a couple due to the additional CRC-16 bit stuffing
         * bits we haven't accounted for. (These are guaranteed to cause less than one
         * byte's worth of difference in the calculation if we do this right.)
         *
         * The full offset calculation is:
         *
         *    headerBits   = 16
         *    crcBits      = 16
         *    roundingBits = 4
         *
         *    offset = (headerBits + crcBits + roundingBits) * 8 - 4 = 284
         *
         * new offset for 80MHz clock   6 2/3 clocks per bit
         *    offset = (headerBits + crcBits + roundingBits) * (20/3) - 4 = 236
         */

        pc = get_port_connection();
        if (pc == PORTC_LOW_SPEED) {
            result = request_data_in(token, txrxFlag, togglePtr, tokenRetries);
            result = ((result - rx2_sop) - 1916) * 3;
            result /= 20;
            result /= 8;
        }
        else {
            //result = ((request_data_in(token, txrxFlag, togglePtr, tokenRetries) - rx2_sop - 236) * 3) / 20;
            result = request_data_in(token, txrxFlag, togglePtr, tokenRetries);
            result = ((result - rx2_sop) - 236) * 3;
            result /= 20;
        }

        if (result <= 0) {
            result = 0;    // Zero-length packet. Device ended the transfer early
        }
        else if (length <= 0) {
            result = 0;    // We don't want the data, the caller is just checking for a good PID. We're done.
        }
        else {
            // The packet wasn't zero-length. Figure out how many bytes it was while
            // we were decoding. The parameters for OP_RX_DATA_PTR are very counterintuitive,
            // see the assembly code for a complete description. After the call, packetLen
            // contains the number of actual bytes stored.
            sync();
            txc_result = ((length - 1) << 16) | result;
            command(OP_RX_DATA_PTR, (uint32_t)buffer);
            sync();
            result = (txc_result & 0xFFFF) - (uint32_t)buffer;

            // We currently can't tell whether RX_DATA_PTR ended because it hit the
            // bit limit or because it hit the byte limit. If it hit the bit limit, all
            // is well and we should be receiving the whole packet. But if it hit the byte
            // limit, this is actually a babble condition. We stopped short, and the receive
            // buffer isn't pointing at the actual CRC.
            //
            // This is why E_CRC can mean either a CRC error or a babble error.

            command(OP_RX_CRC16, 0);
            sync();
            if (txc_result && !(debugFlags & DEBUGFLAG_NO_CRC)) {
                result = E_CRC;
            }
        }

        if (result >= 0) {
            // Success. ACK (if we haven't already) and get out.
            if (txrxFlag != TXRX_TX_RX_ACK) {
                request_data_in(token, TXRX_TX_RX_ACK, togglePtr, tokenRetries);
            }
            return result;
        }
    }
    // Out of CRC retries
    return -1;
}

/*
 * Enqueue a token in the TX buffer
 */
void USBHost::send_token(int pid, int token, int delayAfter)
{
    command(OP_TX_BEGIN, pid);
    command(OP_TX_DATA_16, token);
    command(OP_TX_END, delayAfter);
}

/*
 * Return the raw 16-bit frame from the last E_PID error
 */
int USBHost::last_pid_error()
{
    return last_pid_err;
}

/* ============================================================================== *
 * Low-level Command Interface                                                    *
 * ============================================================================== */

/*
 * Wait for the driver cog to finish what it was doing.
 */
void USBHost::sync()
{
    while (txc_command) {
        Yield__();
    }
}

/*
 * Asynchronously execute a low-level driver cog command.
 * To save space in the driver cog, the conversion from
 * command ID to address happens here, and we pack the
 * address and 16-bit argument into the command word.
 */
void USBHost::command(int cmd, int arg)
{
    uint32_t _lookup[] = {
        0,
        (uint32_t) & cmd_reset >> 2,
        (uint32_t) & cmd_tx_begin >> 2,
        (uint32_t) & cmd_tx_end >> 2,
        (uint32_t) & cmd_txrx >> 2,
        (uint32_t) & cmd_tx_data_16 >> 2,
        (uint32_t) & cmd_tx_data_ptr >> 2,
        (uint32_t) & cmd_tx_crc16 >> 2,
        (uint32_t) & cmd_rx_pid >> 2,
        (uint32_t) & cmd_rx_data_ptr >> 2,
        (uint32_t) & cmd_rx_crc16 >> 2,
        (uint32_t) & cmd_sof_wait >> 2
    };
    int packed;

    packed = _lookup[cmd] | (arg << 16);
    sync();
    txc_command = packed;
}

int USBHost::command_result()
{
    sync();
    return txc_result;
}

void USBHost::command_extra(int arg)
{
    sync();
    txc_result = arg;
}

/*
 * TERMS OF USE: MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

