/*
 * This is a software implementation of a simple full-speed
 * (12 Mb/s) USB 1.1 host controller for the Parallax Propeller.
 *
 * Based on original code by M. Elizabeth Scott
 * https://github.com/scanlime/propeller-usb-host
 *
 * Ported to C / C++ by Marco Maccaferri.
 *
 * Copyright (c) 2010-2016 M. Elizabeth Scott <beth@scanlime.org>
 * Copyright (c) 2020 M. Maccaferri <macca@maccasoft.com>
 */

#ifndef USBHOST_H_
#define USBHOST_H_

#include <propeller.h>
#include <stdint.h>

#include "USB.h"

extern uint32_t _load_start_cog_usb_host_fs_controller[];
extern uint32_t _load_start_cog_usb_host_fs_receiver[];

extern uint32_t * sof_deadline asm("sof_deadline");
extern uint32_t * txp_portc asm("txp_portc");
extern uint32_t * txp_result asm("txp_result");
extern uint32_t * txp_rxdone asm("txp_rxdone");
extern uint32_t * txp_rx1_time asm("txp_rx1_time");
extern uint32_t * txp_rxbuffer asm("txp_rxbuffer");

extern uint32_t * rx1p_done asm("rx1p_done");
extern uint32_t * rx1p_buffer asm("rx1p_buffer");
extern uint32_t * rx1p_sop asm("rx1p_sop");
extern uint32_t * rx1p_portc asm("rx1p_portc");

extern uint32_t cmd_reset asm("cmd_reset");
extern uint32_t cmd_tx_begin asm("cmd_tx_begin");
extern uint32_t cmd_tx_end asm("cmd_tx_end");
extern uint32_t cmd_txrx asm("cmd_txrx");
extern uint32_t cmd_tx_data_16 asm("cmd_tx_data_16");
extern uint32_t cmd_tx_data_ptr asm("cmd_tx_data_ptr");
extern uint32_t cmd_tx_crc16 asm("cmd_tx_crc16");
extern uint32_t cmd_rx_pid asm("cmd_rx_pid");
extern uint32_t cmd_rx_data_ptr asm("cmd_rx_data_ptr");
extern uint32_t cmd_rx_crc16 asm("cmd_rx_crc16");
extern uint32_t cmd_sof_wait asm("cmd_sof_wait");

typedef uint8_t  UCHAR;
typedef uint16_t USHORT;

class USBHost {
public:
    USBHost();

    int start();

    void frame_wait(int count);
    int  get_port_connection();
    int  enumerate();

    int  configure();
    int  unconfigure();

    int  read_configuration(int index);
    int  vendor_id();
    int  product_id();

    void clear_halt(uint8_t * epd);

    uint8_t * next_descriptor(uint8_t * ptrIn);
    uint8_t * next_header_match(uint8_t * ptrIn, uint16_t header);
    uint8_t * first_interface();
    uint8_t * next_interface(uint8_t * curIf);
    uint8_t * next_endpoint(uint8_t * curIf);
    uint8_t * find_interface(int ifClass);

    void device_reset();
    int  device_address();

    int  control_write(int req, int value, int index, uint8_t * buffer_ptr, int length);
    int  control_read(int req, int value, int index, uint8_t * buffer_ptr, int length);
    int  control_raw(int token, uint8_t * buffer_ptr, int length);

    int  interrupt_read(uint8_t * epd, uint8_t * buffer, int length);

    void bulk_write(uint8_t * epd, uint8_t * buffer, int length);
    int  bulk_read(uint8_t * epd, uint8_t * buffer, int length);

    void reset_endpoint_toggle();

    int  data_in(int token, uint8_t * buffer, int length, int maxPacketLen, uint8_t * togglePtr, int txrxFlag, int tokenRetries, int crcRetries);
    int  write_data(int pid, int token, uint8_t *buffer, int length, uint8_t * togglePtr, int retries);
    int  read_data_in(int token, uint8_t * buffer, int length, uint8_t * togglePtr, int txrxFlag, int tokenRetries, int crcRetries);
    int  request_data_in(int token, int txrxFlag, uint8_t * togglePtr, int retries);

    void send_token(int pid, int token, int delayAfter);
    int  last_pid_error();

    int  control(int req, int value, int index);

    /* ==============================================================================
     * Low-level Command Interface
     * ==============================================================================
     */

    void sync();
    void command(int cmd, int arg);
    int  command_result();
    void command_extra(int arg);

public:
    volatile int32_t cog1;
    volatile int32_t cog2;

    volatile uint32_t txc_command;   // Command buffer: [23:16]=arg, [15:0]=code ptr
    volatile uint32_t rx1_time;      // Trigger time for RX1 cog
    // uint32_t rx2_time;               // Trigger time for RX2 cog
    volatile uint32_t rx2_sop;       // Start of packet, calculated by RX2
    volatile uint32_t txc_result;

    volatile uint16_t buf_dd;        // Device descriptor buffer pointer
    volatile uint16_t buf_cfg;       // Configuration descriptor buffer pointer
    volatile uint16_t buf_setup;     // SETUP packet buffer pointer
    volatile uint16_t last_pid_err;  // Details from the last E_PID error

    volatile uint8_t isRunning;
    volatile uint8_t portc;          // Port connection status
    volatile uint8_t rxdone;
    volatile uint8_t debugFlags;

    uint8_t * alloc_buf_dd;
    uint8_t * alloc_buf_cfg;
    uint8_t * alloc_buf_setup;
    uint8_t * alloc_buf_rtx;
};

#endif /* USBHOST_H_ */
