/*
 * This is a software implementation of a simple full-speed
 * (12 Mb/s) USB 1.1 host controller for the Parallax Propeller.
 *
 * Based on original code by M. Elizabeth Scott
 * https://github.com/scanlime/propeller-usb-host
 *
 * Ported to C / C++ by Marco Maccaferri.
 *
 * Copyright (c) 2010-2016 M. Elizabeth Scott <beth@scanlime.org>
 * Copyright (c) 2020 M. Maccaferri <macca@maccasoft.com>
 *
 * See end of file for terms of use.
 */


              .pasm
              .compress off

#include "USB.h"

              .section .cog_usb_host_fs_controller, "ax"

              .global cmd_reset
              .global cmd_tx_begin
              .global cmd_tx_end
              .global cmd_txrx
              .global cmd_tx_data_16
              .global cmd_tx_data_ptr
              .global cmd_tx_crc16
              .global cmd_rx_pid
              .global cmd_rx_data_ptr
              .global cmd_rx_crc16
              .global cmd_sof_wait

              .global sof_deadline
              .global txp_portc
              .global txp_result
              .global txp_rxdone
              .global txp_rx1_time
              .global txp_rxbuffer

              .org      0
controller_cog

              '======================================================
              ' Cog Initialization
              '======================================================

              ' Initialize the PLL and video generator for 12 MB/s output.
              ' This sets up CTRA as a divide-by-8, with no PLL multiplication.
              ' Use 2bpp "VGA" mode, so we can insert SE0 states easily. Every
              ' two bits we send to waitvid will be two literal bits on D- and D+.

              ' To start with, we leave the pin mask in vcfg set to all zeroes.
              ' At the moment we're actually ready to transmit, we set the mask.
              '
              ' We also re-use this initialization code space for temporary variables.

tx_count      mov       ctra, ctra_value
t1            mov       frqa, frqa_value
l_cmd         mov       vcfg, vcfg_value
codec_buf     mov       vscl, vscl_value

codec_cnt     call      #enc_reset

              '======================================================
              ' Command Processing
              '======================================================

              ' Wait until there's a command available or it's time to send a SOF.
              ' SOF is more important than a command, but we have no way of ensuring
              ' that a SOF won't need to occur during a command- so the SOF might be
              ' late.

cmdret
              ' Reduce jitter between spin code and the USB packets, for experiments where that matters
              waitvid   v_palette, v_idle
              waitvid   v_palette, v_idle

              wrlong    c_zero, par
              andn      outa, c_00010000

command_loop
              mov       t1, cnt                 ' cnt - sof_deadline, store sign bit
              sub       t1, sof_deadline
              rcl       t1, #1 wc               ' C = deadline is in the future
        if_nc tjz       tx_count, #tx_sof       ' Send the SOF if the buffer is not in use

              rdlong    l_cmd, par wz           ' Look for an incoming command
        if_z  jmp       #command_loop

              movs      :cmdjmp, l_cmd          ' Handler address in low 16 bits
              rol       l_cmd, #16              ' Now parameter is in low 16 bits
:cmdjmp       jmp       #0

              '======================================================
              ' SOF Packets / PORTC Sampling
              '======================================================

              ' If we're due for a SOF and we're between packets,
              ' this routine is called to transmit the SOF packet.
              '
              ' We're allowed to use the transmit buffer, but we must
              ' not return via 'cmdret', since we don't want to clear
              ' our command buffer- if another cog wrote a command
              ' while we're processing the SOF, we would miss it.
              ' So we need to use the lower-level encoder routines
              ' instead of calling other command implementations.
              '
              ' This happens to also be a good time to sample the port
              ' connection status. When the bus is idle, its state tells
              ' us whether a device is connected, and what speed it is.
              ' We need to sample this when the bus is idle, and this
              ' is a convenient time to do so. We can also skip sending
              ' the SOF if the bus isn't in a supported state.

tx_sof
              xor       cmd_sof_wait, c_condition     ' Let an SOF wait through.
                                                      ' (Swap from if_always to if_never)

              mov       t1, ina
              and       t1, #BUS_MASK
              wrbyte    t1, txp_portc                 ' Save idle bus state as PORTC
              cmp       t1, #PORTC_FULL_SPEED wz      ' Z if full speed
        if_z  jmp       #:sof_full

              mov       v_palette,v_palette_low       ' switch the tx speed while time is not critical
              mov       vscl_value,vscl_low           '
              cmp       t1, #PORTC_LOW_SPEED wz       ' Z if low speed
       if_nz  jmp       #:skip                        ' skip if not low speed
              jmp       #:sof_low

:sof_full
              mov       v_palette,v_palette_full      ' switch the tx speed while time is not critical
              mov       vscl_value,vscl_full          '

              call      #encode_sync                  ' SYNC field

              mov       codec_buf, sof_frame          ' PID and Token
              mov       codec_cnt, #24
              call      #encode
:sof_low

              call      #encode_eop                   ' End of packet and inter-packet delay
              call      #encode_idle ' this added to FS too!

              mov       l_cmd, #0                     ' TX only, no receive
              call      #txrx
              jmp       #:skip

              ' LOW speed uses keep-alive pulses instead of SOF
':keepalive
'              cmp       t1, #PORTC_LOW_SPEED wz
'        if_nz jmp       #:skip                        ' Only send keepalive to low-speed devices

'              mov       t1,#100     ' delay
'              djnz      t1,#$                         '

'              mov       t1,#25     ' 4*N + 8 clocks = 2 bit times (N=25)
'              andn      outa, #BUS_MASK
'              or        dira, #BUS_MASK               ' Send keep-alive
'              djnz      t1,#$                         '
'              or        outa, #PORTC_LOW_SPEED        ' Drive idle
'              mov       t1,#25     ' 4*N + 8 clocks = 2 bit times (N=25)
'              djnz      t1,#$                         '
'              andn      dira, #BUS_MASK               ' release bus
'              andn      outa, #BUS_MASK
'              call      #enc_reset                    ' May not be needed, but txrx does it

:skip
              add       sof_deadline, sof_period

              jmp       #command_loop

              '======================================================
              ' OP_TX_BEGIN
              '======================================================

              ' When we begin a packet, we'll always end up generating
              ' 16 bits (8 sync, 8 pid) which will fill up the first long
              ' of the transmit buffer. So it's legal to use tx_count!=0
              ' to detect whether we're using the transmit buffer.

cmd_tx_begin
              call      #encode_sync

              ' Now NRZI-encode the PID field

              mov       codec_buf, l_cmd
              mov       codec_cnt, #8
              call      #encode

              ' Reset the CRC-16, it should cover only data from after the PID.

              mov       enc_crc16, crc16_mask

              jmp       #cmdret

              '======================================================
              ' OP_TX_END
              '======================================================

cmd_tx_end
              call      #encode_eop
:idle_loop
              test      l_cmd, #$1FF wz
        if_z  jmp       #cmdret
              call      #encode_idle
              sub       l_cmd, #1
              jmp       #:idle_loop

              '======================================================
              ' OP_TX_DATA_16
              '======================================================

cmd_tx_data_16
              mov       codec_buf, l_cmd
              mov       codec_cnt, #16
              call      #encode

              jmp       #cmdret

              '======================================================
              ' OP_TX_DATA_PTR
              '======================================================

              ' Byte count in "result", hub pointer in l_cmd[15:0].
              '
              ' This would be faster if we processed in 32-bit
              ' chunks when possible (at least 4 bytes left, pointer is
              ' long-aligned) but right now we're optimizing for simplicity
              ' and small code size.

cmd_tx_data_ptr
              rdlong    t1, txp_result

:loop         rdbyte    codec_buf, l_cmd
              mov       codec_cnt, #8
              add       l_cmd, #1
              call      #encode
              djnz      t1, #:loop

              jmp       #cmdret

              '======================================================
              ' OP_TX_CRC16
              '======================================================

cmd_tx_crc16
              mov       codec_buf, enc_crc16
              xor       codec_buf, crc16_mask
              mov       codec_cnt, #16
              call      #encode

              jmp       #cmdret

              '======================================================
              ' OP_TXRX
              '======================================================

cmd_txrx
              call      #txrx
              jmp       #cmdret

              '======================================================
              ' OP_RESET
              '======================================================

cmd_reset

              andn      outa, #BUS_MASK         ' Start driving SE0
              or        dira, #BUS_MASK

              mov       t1, cnt
              add       t1, reset_period
              waitcnt   t1, #0

              andn      dira, #BUS_MASK         ' Stop driving
              mov       sof_deadline, cnt       ' Ignore SOFs that should have occurred

              jmp       #cmdret

              '======================================================
              ' OP_RX_PID
              '======================================================

              ' Receive a 16-bit word, and reset the CRC-16.
              ' For use in receiving and validating a packet's SYNC/PID header.

cmd_rx_pid
              mov       codec_cnt, #16
              call      #decode
              shr       codec_buf, #16
              wrlong    codec_buf, txp_result

              mov       dec_crc16, crc16_mask   ' Reset the CRC-16

              jmp       #cmdret

              '======================================================
              ' OP_RX_DATA_PTR
              '======================================================

              ' Parameters:
              '   - Hub pointer in the command word
              '   - Maximum raw bit count in result[15:0]
              '   - Maximum byte count in result[31:16]
              '
              ' Returns:
              '   - Final write pointer (actual byte count + original pointer)
              '
              ' Always decodes at least one byte.
              '
              ' This would be faster if we processed in 32-bit
              ' chunks when possible (at least 4 bytes left, pointer is
              ' long-aligned) but right now we're optimizing for simplicity.
              '
              ' If this is modified to operate on 32-bit words in the future,
              ' this optimization must only take effect when the remaining bit
              ' count is high enough that we're guaranteed not to hit the bit
              ' limit during the 32-bit word. The returned actual byte count
              ' MUST have one-byte granularity.
              '
              ' We stop receiving when the byte or bit counts underflow. So both
              ' counts should be one byte under the actual values.

cmd_rx_data_ptr
              rdlong    t1, txp_result          ' Byte/bit count

:loop         mov       codec_cnt, #8           ' One byte at a time
              sub       t1, c_00010000          ' Decrements byte count
              call      #decode                 ' Decrements bit count
              shr       codec_buf, #24          ' Right-justify result
              wrbyte    codec_buf, l_cmd        ' Store result
              add       l_cmd, #1               ' Pointer + 1
              test      t1, c_80008000 wz       ' Detect bit or byte underflow
        if_z  jmp       #:loop

              wrword    l_cmd, txp_result
              jmp       #cmdret

              '======================================================
              ' OP_RX_CRC16
              '======================================================

cmd_rx_crc16
              xor       dec_crc16, crc16_mask   ' Save CRC of payload
              mov       t3, dec_crc16

              mov       codec_cnt, #16
              call      #decode

              shr       codec_buf, #16          ' Justify received CRC
              xor       t3, codec_buf           ' Compare
              wrlong    t3, txp_result          ' and return
              jmp       #cmdret

              '======================================================
              ' OP_SOF_WAIT
              '======================================================

              ' Normally this jumps back to the command loop without
              ' completing the command. In tx_sof, this code is modified
              ' to return exactly once.
              '
              ' (The modification works by patching the condition code on the
              ' first instruction in this routine.)

cmd_sof_wait  jmp       #command_loop
              xor       cmd_sof_wait, c_condition       ' Swap from if_never to if_always
              jmp       #cmdret


              '======================================================
              ' Transmit / Receive Front-end
              '======================================================

txrx
              ' Save the raw transmit length, not including padding,
              ' then pad our buffer to a multiple of 16 (one video word).

              mov       tx_count_raw, tx_count
:pad          test      tx_count, #%1111 wz
        if_z  jmp       #:pad_done
              call      #encode_idle
              jmp       #:pad
:pad_done

              ' Reset the receiver state (regardless of whether we're using it)

              wrbyte    v_idle, txp_rxdone      ' Arbitrary nonzero byte

              rcr       l_cmd, #1 wc            ' C = bit0 = RX Enable

              ' The RX cog clock pin can be reused as a wakeup pin.
              ' This is great because we don't have to calculate the wakup time
              ' and we save power by using waitpeq instead of looping.
              or        dira, tx_ncmask        ' Maybe put this somewhere else
       if_c   or        outa, tx_ncmask        ' Get ready for falling edge later

              ' Right now tx_count_raw is the number of bits
              ' in the packet. Convert it to a loop count we can
              ' use to line up our EOP with the video generator phase.
              ' We must multiply by 1 2/3 for 80MHz
              ' instead of 2 for 96MHz.
              ' approximated by 1 3/4
              ' peak error +1 or -1 instruction

              add       tx_count_raw, #15       ' 0 -> 16
              and       tx_count_raw, #%1111    ' Period is 16 bits
              mov       t1,tx_count_raw
              shr       t1, #1                  '
              add       tx_count_raw,t1         ' 1/2
              shr       t1, #1                  '
              add       tx_count_raw,t1         ' 1/4
              cmp       vscl_value,vscl_low wz
        if_z  shl       tx_count_raw,#3         ' low speed 8 times slower
        if_z  add       tx_count_raw,#20
              ' Transmitter startup: We need to synchronize with the video PLL,
              ' and transition from an undriven idle state to a driven idle state.
              ' To do this, we need to fill up the video generator register with
              ' idle states before setting DIRA and VCFG.
              '
              ' Since we own the bus at this point, we don't have to hurry.

              ' This section was relocated because the extra math involved
              ' for 80MHz operation was running causing the video generator
              ' to run out of data and transmit garbage on the bus.
              ' May not apply with pin wakeup
              ' Make sure there are no more than XX instructions between waitvids.

              mov       vscl, vscl_value                ' Back to normal video speed
              waitvid   v_palette, v_idle
              waitvid   v_palette, v_idle
              movs      vcfg, #BUS_MASK
              or        dira, #DEBUG_TX_MASK | BUS_MASK

              ' Transmit our NRZI-encoded packet.
              '
              ' This loop is optimized to do the last waitvid separately, so
              ' that we don't add any extra instructions between it and the
              ' bus release code below.

              movs      _tx_inst1, #tx_buffer
              shr       tx_count, #4            ' Bits -> words

_tx_loop      sub       tx_count, #1 wz
        if_z  jmp       #_tx_loop_last          ' Stop looping before the last word

_tx_inst1     waitvid   v_palette, 0            ' Output all words except the last one
              add       _tx_inst1, #1
              jmp       #_tx_loop

_tx_loop_last mov       _tx_inst2, _tx_inst1    ' Copy last address
              mov       _tx_inst3, _tx_inst1
    if_c      andn      outa, tx_ncmask         ' Falling edge enables RX cog

              ' The last word is special, since we need to stop driving the bus
              ' immediately after the video generator clocks out our EOP bits.
              ' We've already calculated a loop count for how long to delay
              ' between the waitvid and the end of our EOP- but we need to
              ' special-case 0 so we can stop driving immediately after waitvid.

              tjz       tx_count_raw, #_tx_inst3

_tx_inst2     waitvid   v_palette, 0            ' Output last word
              djnz      tx_count_raw, #$        ' Any tx_count_raw >= 1
              andn      dira, #DEBUG_TX_MASK | BUS_MASK
              jmp       #_tx_release_done

_tx_inst3     waitvid   v_palette, 0            ' tx_count_raw == 0
              andn      dira, #DEBUG_TX_MASK | BUS_MASK


_tx_release_done

              ' As soon as we're done transmitting, switch to a 'turbo' vscl value,
              ' so that after the current video word expires we switch to a faster
              ' clock. This will help us synchronize to the video generator faster
              ' when sending ACKs, decreasing the maximum ACK latency.

              mov       vscl, vscl_turbo

              '======================================
              ' Receiver Controller
              '======================================

        if_nc jmp       #_rx_done                       ' Receiver disabled
              rcr       l_cmd, #1 wc                    ' C = bit1 = ACK Enable

              ' First, wait for an EOP signal. This wait needs to have a timeout,
              ' in case we never receive a packet. It also needs to have low latency,
              ' since we use this timing both to send ACK packets and to calculate
              ' the length of the received packet.

              mov       t1, eopwait_iters
_wait_eop     test      c_bus_mask, ina wz
        if_nz djnz      t1, #_wait_eop
              mov       t3, cnt                         ' EOP timestamp

              mov       t1,#20
              cmp       vscl_value,vscl_low wz
        if_z  djnz      t1,#$         ' low speed delay

              ' The USB spec gives us a fairly narrow window in which to transmit the ACK.
              ' So, to get predictable latency while also keeping code size down, we
              ' use the video generator in a somewhat odd way. Prior to this code,
              ' we set the video generator to run very quickly, so the variation in
              ' waitvid duration is fairly small. After re-synchronizing to the video
              ' generator, we slow it back down and emit a pre-constructed ACK packet.

        if_c  waitvid   v_palette, v_idle                ' Sync to vid gen. at turbo speed
        if_c  mov       vscl, vscl_value                 ' Back to normal speed at the next waitvid
        if_c  waitvid   v_palette, v_ack1                ' Start ACK after a couple idle cycles
        if_c  or        dira, #DEBUG_ACK_MASK | BUS_MASK ' Take bus ownership during the idle
        if_c  waitvid   v_palette, v_ack2                ' Second half of the ACK + EOP
        if_c  waitvid   v_palette, v_idle                ' Wait for the ack to completely finish
        if_c  andn      dira, #DEBUG_ACK_MASK | BUS_MASK ' Release bus

              ' Time-critical work is over. Save the EOP timestamp. The Spin code
              ' will use this value to calculate actual packet length.

              wrlong    t3, txp_result

              ' Now we're just waiting for the RX cog to finish. Poll RX_DONE.
              ' This shouldn't take long, since we already waited for the EOP.
              ' The RX cogs just need to detect the EOP and finish the word
              ' they're on. We'll be conservative and say they need 64 bit
              ' periods (2 full iterations) to do this job. That's 512
              ' clock cycles, or 32 hub windows.

              mov       t1, #1
              shl       t1,#12
              'mov       t1,#32
_rx_wait      rdbyte    t3, txp_rxdone wz
        if_nz djnz      t1, #_rx_wait

              ' If the timeout expired and our RX cogs still aren't done,
              ' we'll manually wake them up by driving a SE1 onto the bus
              ' for a few cycles.

        if_nz or        outa, #BUS_MASK
        if_nz or        dira, #BUS_MASK
              nop
              nop
        if_nz andn      dira, #BUS_MASK
        if_nz andn      outa, #BUS_MASK

              ' Initialize the decoder, load the first long into the shift register.
              ' Add the first zero to the sync byte since the receiver misses it.

                mov     dec_rxbuffer, txp_rxbuffer
                mov     dec_nrzi_cnt, #31        ' 31 bits for first long
                mov     dec_1cnt, #0
                rdlong  dec_nrzi, dec_rxbuffer
                shl     dec_nrzi,#1              ' add the implied zero
                add     dec_rxbuffer, #4

_rx_done
              '======================================
              ' End of Receiver Controller
              '======================================

              call      #enc_reset              ' Reset the encoder too
              movs      vcfg, #0                ' Disconnect vid gen. from outputs

txrx_ret      ret


              '======================================================
              ' NRZI Encoding and Bit Stuffing
              '======================================================

              ' Encode (NRZI, bit stuffing, and store) up to 32 bits.
              '
              ' The data to be encoded comes from codec_buf, and codec_cnt
              ' specifies how many bits we shift out from the LSB side.
              '
              ' For both space and time efficiency, this routine is also
              ' responsible for updating a running CRC-16. This is only
              ' used for data packets- at all other times it's quietly
              ' ignored.
encode
              rcr       codec_buf, #1 wc

              ' Update the CRC16.
              '
              ' This is equivalent to:
              '
              '   condition = (input_bit ^ (enc_crc16 & 1))
              '   enc_crc16 >>= 1
              '   if condition:
              '     enc_crc16 ^= crc16_poly

              test      enc_crc16, #1 wz
              shr       enc_crc16, #1
    if_z_eq_c xor       enc_crc16, crc16_poly

              ' NRZI-encode one bit.
              '
              ' For every incoming bit, we generate two outgoing bits;
              ' one for D- and one for D+. We can do all of this in three
              ' instructions with SAR and XOR. For example:
              '
              '   Original value of tx_reg:        10 10 10 10
              '   After SAR by 2 bits:          11 10 10 10 10
              '     To invert D-/D+, flip MSB:  01 10 10 10 10
              '    (or)
              '     Avoid inverting by flipping
              '     the next highest bit:       10 10 10 10 10
              '
              ' These two operations correspond
              ' to NRZI encoding 0 and 1, respectively.

              sar       enc_nrzi, #2
        if_nc xor       enc_nrzi, c_80000000     ' NRZI 0
        if_c  xor       enc_nrzi, c_40000000     ' NRZI 1


              ' Bit stuffing: After every six consecutive 1 bits, insert a 0.
              ' If we detect that bit stuffing is necessary, we do the branch
              ' after storing the original bit below, then we come back here to
              ' store the stuffed bit.

        if_nc mov       enc_1cnt, #6 wz
        if_c  sub       enc_1cnt, #1 wz
enc_bitstuff_ret

              ' Every time we fill up enc_nrzi, append it to tx_buffer.
              ' We use another shift register as a modulo-32 counter.

              ror       enc_nrzi_cnt, #1 wc
              add       tx_count, #1
encode_ptr
        if_c  mov       0, enc_nrzi
        if_c  add       encode_ptr, c_dest_1

              ' Insert the stuffed bit if necessary

        if_z  jmp       #enc_bitstuff

              djnz      codec_cnt, #encode
encode_ret    ret

              ' Handle the relatively uncommon case of inserting a zero bit,
              ' for bit stuffing. This duplicates some of the code from above
              ' for NRZI-encoding the extra bit. This bit is *not* included
              ' in the CRC16.

enc_bitstuff  sar       enc_nrzi, #2
              xor       enc_nrzi, c_80000000
              mov       enc_1cnt, #6 wz
              jmp       #enc_bitstuff_ret       ' Count and store this bit


              '======================================================
              ' Encoder / Transmitter Reset
              '======================================================

              ' (Re)initialize the encoder and transmitter registers.
              ' The transmit buffer will now be empty.

enc_reset     mov       enc_nrzi, v_idle
              mov       enc_nrzi_cnt, enc_ncnt_init
              mov       enc_1cnt, #0
              mov       tx_count, #0
              movd      encode_ptr, #tx_buffer
enc_reset_ret ret


              '======================================================
              ' Low-level Encoder
              '======================================================

              ' The main 'encode' function above is the normal case.
              ' But we need to be able to encode special bus states too,
              ' so these functions are slower but more flexible encoding
              ' entry points.
              '

              ' Check whether we need to store the contents of enc_nrzi
              ' after encoding another bit-period worth of data from it.
              ' This is a modified version of the tail end of 'encode' above.

encode_store
              mov       :ptr, encode_ptr
              ror       enc_nrzi_cnt, #1 wc
              add       tx_count, #1
:ptr    if_c  mov       0, enc_nrzi
        if_c  add       encode_ptr, c_dest_1
encode_store_ret ret

              ' Raw NRZI zeroes and ones, with no bit stuffing

encode_raw0
              sar       enc_nrzi, #2
              xor       enc_nrzi, c_80000000
              mov       enc_1cnt, #0
              call      #encode_store
encode_raw0_ret ret

encode_raw1
              sar       enc_nrzi, #2
              xor       enc_nrzi, c_40000000
              call      #encode_store
encode_raw1_ret ret

              ' One cycle of single-ended zero.

encode_se0
              shr       enc_nrzi, #2
              call      #encode_store
encode_se0_ret ret

              ' One cycle of idle bus (J state).

encode_idle
              shr       enc_nrzi, #2
              xor       enc_nrzi, c_80000000
              call      #encode_store
encode_idle_ret ret

              ' Append a raw SYNC field
encode_sync
              mov       t1, #7
:loop         call      #encode_raw0
              djnz      t1, #:loop
              call      #encode_raw1
encode_sync_ret ret

              ' Append a raw EOP.
              '
              ' Note that this makes sure we have at least one idle
              ' bit after the SE0s, but we'll probably have more due
              ' to the padding and bus-release latency in the transmitter.
encode_eop
              call      #encode_se0
              call      #encode_se0
              call      #encode_idle
encode_eop_ret ret


              '======================================================
              ' NRZI Decoder / Bit un-stuffer
              '======================================================

              ' Decode (retrieve, NRZI, bit un-stuff) up to 32 bits.
              '
              ' The NRZI decoding is now done in the rx cog.
              ' We still need to remove stuffed bits and calculate CRC-16.
              '
              ' As with encoding, we also run a CRC-16 here, since it's
              ' a convenient place to do so.
              '
              ' For every raw bit we consume, subtract 1 from t1.
              ' This is used as part of the byte/bit limiting for rx_data_ptr.

decode
              sub       t1, #1

              ' Variables redefined since previous versions:
              ' dec_nrzi_cnt   = number of bits remaining in dec_nrzi
              ' dec_nrzi       = shift register

              rcr       dec_nrzi, #1 wc        ' shift new bit to c
              sub       dec_nrzi_cnt, #1 wz    ' count bits remaining
        if_z  rdlong    dec_nrzi, dec_rxbuffer ' reload if necessary
        if_z  add       dec_rxbuffer, #4
        if_z  mov       dec_nrzi_cnt, #30

              cmp       dec_1cnt, #6 wz         ' Skip stuffed bits
        if_z  mov       dec_1cnt, #0
        if_z  jmp       #decode

              rcr       codec_buf, #1            ' add new data at bit 31
              test      codec_buf, c_80000000 wz ' Move decoded bit to Z


    if_nz     add       dec_1cnt, #1            ' Count consecutive '1' bits
    if_z      mov       dec_1cnt, #0

              ' Update our CRC-16. This performs the same function as the logic
              ' in the encoder above, but it's optimized for our flag usage.

              shr       dec_crc16, #1 wc          ' Shift out CRC LSB into C
    if_z_eq_c xor       dec_crc16, crc16_poly

              djnz      codec_cnt, #decode
decode_ret    ret


              '======================================================
              ' Data
              '======================================================

' Parameters that are set up by Spin code prior to cognew()

sof_deadline  long      0
txp_portc     long      0
txp_result    long      0
txp_rxdone    long      0
txp_rx1_time  long      0
'txp_rx2_time  long      0
txp_rxbuffer  long      0

' Constants

c_zero        long      0
c_40000000    long      $40000000
c_80000000    long      $80000000
c_00010000    long      $00010000
c_80008000    long      $80008000
c_dest_1      long      1 << 9
c_condition   long      %000000_0000_1111_000000000_000000000
c_bus_mask    long      BUS_MASK

reset_period  long      80_000 * 10

frqa_value    long      $13333333                       ' 6MHz
ctra_value    long      (%00001 << 26) | (%111 << 23)   ' PLL x16  96MHz
vcfg_value    long      (%011 << 28)                    ' Unpack 2-bit -> 8-bit
vscl_value    long      (8 << 12) | (8 * 16)            ' Normal 8 clocks per pixel, 16 bits/frame
vscl_turbo    long      (1 << 12) | (1 * 16)            ' 1 clock per pixel, 2 bit times/frame
v_palette     long      (BUS_MASK << 24) | (STATE_J << 16) | (STATE_K) << 8

vscl_low      long      ((8*8) << 12) | ((8*8) * 16)
vscl_full     long      (8 << 12) | (8 * 16)
v_palette_low  long      (BUS_MASK << 24) | (STATE_J << 8) | (STATE_K) << 16 ' Swap D+, D-
v_palette_full long      (BUS_MASK << 24) | (STATE_J << 16) | (STATE_K) << 8

v_idle        long      %%2222_2222_2222_2222
tx_ncmask     long      1<<USBNC

' Pre-encoded ACK sequence:
'
'    SYNC     ACK      EOP
'    00000001 01001011
'    KJKJKJKK JJKJJKKK 00JJ
'
'    waitvid: %%2200_1112_2122_1121_2121
'
' This encoded sequence requires 40 bits.
' We encode this in two logs, but we don't start
' it immediately. Two reasons:
'
'   - We enable the output drivers immediately
'     after sync'ing to the video generator, So
'     there is about a 1/2 bit delay between
'     starting to send this buffer and when we
'     actually take control over the bus.
'
'   - We need to ensure a minimum inter-packet
'     delay between EOP and ACK.
'
'
'     (Currently we wait 4 bit periods.
'      This may need to be tweaked)

v_ack1        long      %%2122112121212222
v_ack2        long      %%2222222222001112

enc_ncnt_init long      $8000_8000                      ' Shift reg as mod-16 counter

crc16_poly    long      $a001                           ' USB CRC-16 polynomial
crc16_mask    long      $ffff                           ' Init/final mask

' How long will we wait for an EOP, during receive?
'
' This is a two-instruction (test / djnz) loop which takes 8 cycles. So each
' iteration is one bit period, and this is really a count of the maximum number
' of bit periods that could exist between the end of the transmitted packet
' and the EOP on the received packet. So it must account for the max size of the
' receive buffer, plus an estimate of the max inter-packet delay.

eopwait_iters   long    ((RX_BUFFER_WORDS * 32) + 128 + 2000)

' We try to send SOFs every millisecond, but other traffic can preempt them.
' Since we're not even trying to support very timing-sensitive devices, we
' also send a fake (non-incrementing) frame number.

sof_frame       long    %00010_00000000000_1010_0101    ' SOF PID, Frame 0, valid CRC5
sof_period      long    80_000                          ' 80 MHz, 1ms

' Encoder only
enc_nrzi        long    0                               ' Encoded NRZI shift register
enc_1cnt        long    0
enc_nrzi_cnt    long    0                               ' Cyclic bit counter
enc_crc16       long    0

' Decoder only
dec_nrzi        long    0                               ' Encoded NRZI shift register
dec_nrzi_cnt    long    0                               ' Bit counter
dec_1cnt        long    0
dec_rxbuffer    long    0
dec_crc16       long    0

tx_count_raw    long    0
t3              long    0

led_tmp         long    0 ' FIXME
tx_buffer

                fit
                
/*
 * TERMS OF USE: MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

