/*
 * This is a software implementation of a simple full-speed
 * (12 Mb/s) USB 1.1 host controller for the Parallax Propeller.
 *
 * Based on original code by M. Elizabeth Scott
 * https://github.com/scanlime/propeller-usb-host
 *
 * Ported to C / C++ by Marco Maccaferri.
 *
 * Copyright (c) 2010-2016 M. Elizabeth Scott <beth@scanlime.org>
 * Copyright (c) 2020 M. Maccaferri <macca@maccasoft.com>
 *
 * See end of file for terms of use.
 */


                .pasm
                .compress off

#include "USB.h"

                .section .cog_usb_host_fs_receiver, "ax"

                .global rx1p_done
                .global rx1p_buffer
                .global rx1p_sop
                .global rx1p_portc


                .org    0
rx_cog_1
                wrlong  rx1_zero, par           ' Notify Start() that we're running.

_restart
                ' Initial conditions for counters
                mov     frqb, #0                ' PHSB is set later
                mov     ctrb, ctr_data_fs       ' accumulate when sampling pulse and data

                ' Initialize CTRA. We must keep the USBNC pin low now,
                ' or we will miss our wakeup signal from the TX cog.
                mov     frqa, #0
                mov     phsa, phs_usb
                mov     ctra, ctr_usb           ' duty output on USBNC pin
                mov     dira, rx1_dira

                ' FIXME We count by longs, not bytes.
                ' The math works out anyway.
                mov     rx1_iters, #RX_BUFFER_WORDS

                ' The sampling loop is pipelined so the
                ' first "received raw data" is actually the
                ' initial conditions set here. We define this
                ' to be (1<<29) to feed the NRZI decoder the
                ' initial one that our sampling loop missed.

                ' The NRZI decoder outputs undefined data for
                ' the first long because oldraw is not initialized.
                ' That's ok, we don't even copy it to hub ram.

                ' The first raw data is (newphs-oldphs).
                ' We set oldphs to zero because it's convenient.

                mov     newphs, d_first
                mov     phsb, newphs            ' load the counter with our known value
                mov     oldphs, #0
                mov     dptr, #databuf

                ' A rising edge from the TX cog is our wakeup signal.
                waitpeq   rx1_nc, rx1_nc

                ' Check what speed we need to receive.
                rdbyte  t2, rx1p_portc
                test    t2, rx1_dplus wz
    if_z        jmp     #_start_ls


_rx1_start_fs
                waitpne rx1_nc,rx1_nc           ' trigger on falling edge

                ' Now synchronize to the beginning of the next packet.
                ' We sample only D- in the receiver. If we time out,
                ' the controller cog will artificially send a SE1
                ' to bring us out of sleep. (We'd rather not send a SE0,
                ' since we may inadvertently reset the device.)

                ' The receiver must start in the narrow window
                ' between the SE0 sent at end of packet and sync byte
                ' at the beginning of the received packet.

                ' The TX cog triggers us a little early
                ' so we wait for SE0 before activating the receiver.
                waitpeq rx1_zero, rx1_dmask     ' wait for EOP

                ' We should have time to do this here since USB spec
                ' require minimum of 2 bit times before next packet.
                mov     frqb, #1 '0             ' get timer ready to receive data

                waitpne rx1_zero, rx1_dminus    ' sync to rising edge (K)
                mov     frqa, frq_usb           ' turn on sampling clock
                mov     rx1_cnt, cnt            ' save SOP time
                shl     frqb, #1 '1
                sub     rx1_cnt, #4             ' adjust SOP timestamp
_sample_loop
                shl     frqb, #1 '2
                ' quantum break
                shl     frqb, #1 '3
                mov     newraw, newphs          ' set up to subtract
                shl     frqb, #1 '4
                sub     newraw, oldphs wz       ' extract new data, detect pEOP
                shl     frqb, #1 '5
                ' quantum break
                shl     frqb, #1 '6
    if_z        jmp     #_rx_done               ' exit quickly on pEOP
                shl     frqb, #1 '7
                movd    _wr, dptr
                shl     frqb, #1 '8
                ' quantum break
                shl     frqb, #1 '9
                add     dptr, #1
                shl     frqb, #1 '10
                mov     datalong, newraw        ' start of nrzi decoder
                shl     frqb, #1 '11
                ' quantum break
                shl     frqb, #1 '12
                shr     oldraw, #29             ' get last bit of old data
                shl     frqb, #1 '13
                and     oldraw, #1              ' destructively
                shl     frqb, #1 '14
                ' quantum break
                shl     frqb, #1 '15
                shl     datalong, #1            ' make room for last bit
                shl     frqb, #1 '16
                add     datalong, oldraw        ' add the last bit on
                shl     frqb, #1 '17
                ' quantum break
                shl     frqb, #1 '18
                xor     datalong, newraw        ' find transitions
                shl     frqb, #1 '19
                xor     datalong,data_mask      ' invert
                shl     frqb, #1 '20
                ' quantum break
                shl     frqb, #1 '21
                nop
                shl     frqb, #1 '22
                nop
                shl     frqb, #1 '23
                ' quantum break
                shl     frqb, #1 '24
                nop
                shl     frqb, #1 '25
_wr             mov     0, datalong             ' store data
                shl     frqb, #1 '26
                ' quantum break
                shl     frqb, #1 '27
                mov     oldraw, newraw
                shl     frqb, #1 '28
                mov     oldphs, newphs
                shl     frqb, #1 '29
                ' quantum break
                mov     frqb, #1 '0
                mov     newphs, phsb            ' read data, bit 0 not yet included
                shl     frqb, #1 '1
                djnz    rx1_iters, #_sample_loop
                ' shl     frqb, #1 '2  after jump

                ' We don't have time to read and clear phsb without
                ' loosing a bit. Thankfully, addition is reversible
                ' so all we need to do is read.

_rx_done
                mov     frqa, #0                ' stop sample clock

                movd    _zwr, dptr              ' set D to add terminating zero to buffer
                mov     frqb, #0                ' stop accum, also a delay between movd and mov
_zwr            mov     0, #0                   ' write zero at end
                                                ' databuf[0] is written with dummy data
                                                ' databuf[1] contains first real data, [31:30] undefined, [29:0] NRZI decoded
                                                ' databuf[n] is written with 0 on pEOP

                jmp #_upload

_start_ls

                movd    _lswr, #databuf         ' reset data pointer
                mov     bitpos,rx1_dplus        ' initial state for waitpxx
                or      bitpos,rx1_nc

                waitpne rx1_nc,rx1_nc           ' trigger on falling edge`
                waitpeq rx1_zero, rx1_dmask     ' wait for EOP

                mov     frqa, #1                ' watchdog timer on
                mov     phsa, phs_wdt
                mov     ctra, ctr_wdt           ' NCO output on NC pin

                waitpne rx1_zero,rx1_dplus      ' wait for SOP (K)

                ' alternate SOP detector with timeout
                ' mov       loopcnt,#1 ' eopwait_iters
                ' shl       loopcnt,#20
' :wait_sop        test      mask, ina wz         ' Wait for a one on
'        if_z     djnz      loopcnt, #:wait_sop  ' D+

                mov     rx1_cnt, cnt            ' save SOP
                mov     oldphs,rx1_cnt          ' also use it as first edge

                ' We use the Z flag and C flag parity calculation to
                ' detect EOP and timeout in the same test.
                '
                ' If we receive a SE1 at the exact same time
                ' as the timer runs out, we will miss it. (Unlikely)
                '
                '  NC       USB      Z      C
                ' ---------------------------------
                '   1       SE0      0      1    Exit due to SE0
                '   1        D+      0      0    Normal receving case
                '   1        D-      0      0    Normal receving case
                '   1       SE1      0      1    Exit due to SE1
                '   0       SE0      1      0    Exit due to timeout and SE0
                '   0        D+      0      1    Exit due to timeout
                '   0        D-      0      1    Exit due to timeout
                '   0       SE1      0      0    Fail to exit !!!

                ' We need waitpxx to continue on SE0
                ' in order to detect SE0 after J.
_ls_loop
                waitpne bitpos, rx1_wmask       ' 1
                mov     newphs, cnt             ' 2
                xor     bitpos, rx1_dmask       ' 3 swap polarity for next edge
                mov     newraw, newphs          ' 4
                test    rx1_wmask, ina  wc,wz   ' 5
                sub     newraw, oldphs          ' 6
_lswr           mov     0, newraw               ' 7 store data
    if_c_or_z   jmp     #_ls_eop                ' 8
                add     _lswr, d_by_one         ' 9
                mov     oldphs, newphs          '10
                djnz    rx1_iters, #_ls_loop    '11

_ls_eop
                ' If we needed the EOP timestamp we could get it from newphs.

                ' Retrieve the current dptr from our write instruction
                mov     dptr, _lswr
                shr     dptr, #9
                and     dptr, #$1FF

                mov     frqa, #0                ' Stop counter
                mov     phsa, #0                ' Set the counter output pin to low
                mov     ctra, ctr_usb           ' ctr_hub, for upload (same as ctr_usb)


                ' The output of the low speed receiver needs some post-processing.
                ' In the future we could do unstuffing here.

                ' oldphs = input  clocks between edges
                ' newphs = used to count how many bits to output for each edge
                ' newraw = output data buffer
                ' ocnt   = output bits remaining in long

                mov     loopcnt, dptr           ' calculate how many
                sub     loopcnt, #databuf       ' longs we received
                add     loopcnt, #2             ' index adjustment   Don't fully understand yet

                movs    _lsread, #(databuf+0)   ' reset read address
                mov     dptr, #(databuf+1)      ' reset write address
                cmp     rx1_zero, #$1FF   wc    ' set carry
                jmp     #_lsinit                ' rest of init in loop

_lsread         mov     newphs, databuf+1       ' read new data
                add     _lsread, #1             ' S++
                sub     loopcnt, #1   wz        ' limit input length
    if_z        jmp     #_lsrx_done

                ' 53 1/3 clocks per bit nominal at 80 MHz
                ' Using 53 shifts the thresholds a little
                ' off center, but we should still be able
                ' to tolerate 5% clock error with 7 consecutive ones.
                ' USB spec is 1.5% for low speed devices.

                sub     newphs, #26   wc
'    if_c        jmp     #_lsread                ' optional noise/error catch

                cmp     newphs, #450  wc        ' Limit the number of bits in each run
    if_nc       jmp     #_lsrx_done             ' to avoid overflow. This would be a
                                                ' bit stuffing error or other fault.

_lswloop        sub     newphs, #53   wc        ' C=1 for last bit of the loop.
    if_nc       or      newraw, bitpos          ' That means write a zero.
                ' Future: skip the next 2 lines for stuffed bits
                shl     bitpos, #1
                sub     lsrx_ocnt, #1 wz        ' count bits remaining
    if_nz       jmp     #_samelong
_lswrite        mov     0, newraw
                add     dptr, #1                ' increment address
_lsinit         movd    _lswrite, dptr
                mov     lsrx_ocnt, #30
                mov     bitpos, #1              ' reset output shift reg
                mov     newraw, #0
_samelong
    if_nc       jmp     #_lswloop
                jmp     #_lsread

_lsrx_done      movd    _lswrite2, dptr
                nop
_lswrite2       mov     0, newraw

_upload         ' Data received, now copy the data to hub ram.
                '
                mov     loopcnt, dptr           ' calculate how many
                sub     loopcnt, #databuf       ' longs we received
                ' loopcnt is number of longs including terminating 0

                movd    _hublp, #databuf        ' set cog start address
                add     _hublp, d_by_one        ' skip dummy long at start
                mov     ctrb, ctr_clock         ' accumulate with sampling pulse only

                ' save the SOP timestamp  and synchronize with hub
                wrlong  rx1_cnt, rx1p_sop
                mov     phsa, phs_hub           ' initial offset for
                mov     phsb, rx1p_buffer       ' set hub start address

                ' Signal that we are done even though
                ' we have yet to copy the data.
                ' The decode cog won't catch us.
                wrbyte  rx1_zero, rx1p_done
                mov     frqb, #4                ' increment by longs
                mov     frqa, frq_hub           ' enable clock to ctrb
                ' Copy the data longs.
_hublp          wrlong  0, phsb                 ' counter increments hub address
                add     _hublp, d_by_one        ' increment cog mem address
                djnz    loopcnt, #_hublp

                mov     frqa, #0                ' stop sample clock
                mov     frqb, #0                ' stop accumulator

        ' Debug code for quickstart LEDs
'       mov cnttmp,databuf+1
   '    mov cnttmp,t2
'      mov t2, rx1p_buffer
'      add t2,#4
'      rdlong cnttmp,t2
   '   mov cnttmp,phsb
'      shr cnttmp,#7
'       and cnttmp,#$ff
'       shl cnttmp,#16
'       mov outa,cnttmp

                jmp     #_restart

rx1_wmask       long    (1<<DPLUS) + (1<<DMINUS) + (1<<USBNC)
rx1_dmask       long    (1<<DPLUS) + (1<<DMINUS)
rx1_dplus       long    1 << DPLUS
rx1_dminus      long    1 << DMINUS
rx1_zero        long    0
rx1_nc          long    1 << USBNC
rx1_dira        long    (1<<USBNC) + ($ff<<16) '+ (1<<7)

ctr_wdt         long    (%00100_000 << 23) + (USBNC) 'nco out
phs_wdt         long    -(53*130)    ' 130 bits is max LS packet + gap

ctr_usb         long    (%00110_000 << 23) + (USBNC) 'duty out

' Adjusting the starting phase may require the
' sampling loop to be modified.
' USB start phase was adjusted using scope to place
' the sampling pulses in the center of the USB bits.

phs_usb         long    $2666_6666*5            ' starting phase
frq_usb         long    $2666_6666              ' 12 MHz pulses for USB

phs_hub         long    $8000_0000              ' Not sure if this is critical
frq_hub         long    $1000_0000              ' 1/16 clock rate

' ctr_data  is used to sample USB data
' ctr_clock is used to increment hub address

ctr_data_fs     long    (%11000_000 << 23) + (USBNC) + (DMINUS<<9)
ctr_clock       long    (%11010_000 << 23) + (USBNC)

d_by_one        long    (1<<9)
data_mask       long    $3FFF_FFFF              ' bits 29-0
d_first         long    (1<<29)                 ' initial condition for NRZI decoder

' Parameters that are set up by Spin code prior to cognew()

rx1p_done       long    0
rx1p_buffer     long    0
rx1p_sop        long    0
rx1p_portc      long    0                       ' we check this for speed data

'rx1_buffer      long    0                        ' we store this in phsb now
rx1_cnt         long    0                       ' stores SOP timestamp
rx1_iters       long    0
t2              long    0
loopcnt         long    0
newphs          long    0
oldphs          long    0
'lsrx_icnt       long    0
lsrx_ocnt       long    0
dptr            long    0
newraw          long    0
oldraw          long    0
datalong        long    0
cnttmp          long    0
'lsrx_cnt        long    0
'iloop           long    0
'oloop           long    0
'sample0         long    0
'sample1         long    0
'sample2         long    0
bitpos          long    0
'lrx_shift       long    0
databuf         long    0

                fit

/*
 * TERMS OF USE: MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

